from django.contrib.auth import authenticate, login as django_login, logout as django_logout, get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.validators import EmailValidator
from django.contrib.auth.signals import user_logged_in, user_logged_out
from django.utils import timezone
from django.core.exceptions import ValidationError
from django.views.decorators.csrf import ensure_csrf_cookie
from ninja import Router
from ninja.responses import Response

from knightauth.models import get_token_model
from knightauth.schemas import LoginIn, ErrorOut, LoginSuccessOut, UserRegisterSchema
from knightauth.settings import knight_auth_settings

token_auth_router = Router()


@token_auth_router.post(
    "login",
    auth=None,
    response={200: LoginSuccessOut, frozenset({401, 403}): ErrorOut},
    url_name="token_login"
)
def token_login(request, payload: LoginIn):
    if knight_auth_settings.TOKEN_LIMIT_PER_USER is not None:
        now = timezone.now()
        token_count = (
            request
            .user
            .auth_token_set
            .filter(created__gte=now)
            .count()
        )

        if token_count >= knight_auth_settings.TOKEN_LIMIT_PER_USER:
            return 403, {"message": "Maximum amount of tokens allowed per user exceeded."}

    user = authenticate(request, **payload.dict())

    if user is None:
        return 401, {"message": "Invalid credentials"}

    instance, token = (
        get_token_model()
        .objects
        .create(
            user=user,
            prefix=knight_auth_settings.TOKEN_PREFIX,
            expiry=knight_auth_settings.TOKEN_TTL
        )
    )

    user_logged_in.send(sender=user.__class__, request=request, user=user)

    return 200, {
        "token": token,
        "expiry": instance.expiry
    }


@token_auth_router.post("logout", response={204: None, 400: ErrorOut}, url_name="token_logout")
def token_logout(request):
    if not hasattr(request, "_auth") or not request._auth:
        return 400, {
            "message": ""
                       "Attempting to log out using an authentication method different from the one used for login."
                       "Consider using session authentication instead"
                       ""
        }

    request._auth.delete()
    user_logged_out.send(sender=request.user.__class__, request=request, user=request.user)

    return 204, None


@token_auth_router.post("logoutall", response={204: None, 400: ErrorOut}, url_name="token_logoutall")
def token_logout_all(request):
    if not request.auth.auth_token_set.exists():
        return 400, {
            "message": ""
                       "Attempting to log out using an authentication method different from the one used for login."
                       "Consider using session authentication instead"
                       ""
        }

    request.auth.auth_token_set.all().delete()
    user_logged_out.send(sender=request.user.__class__, request=request, user=request.user)
    return 204, None


session_auth_router = Router()


@session_auth_router.post("login", auth=None, response={200: None, 400: ErrorOut}, url_name="session_login")
def session_login(request, payload: LoginIn):
    username = payload.username
    password = payload.password

    if username is None or password is None:
        return 400, {"message": "Please provide username and password."}

    user = authenticate(username=username, password=password)

    if user is None:
        return 400, {"message": "Invalid credentials."}

    django_login(request, user)

    return 200, None


@session_auth_router.post("logout", response={200: None}, url_name="session_logout")
def session_logout(request):
    django_logout(request)

    return 200, None


@session_auth_router.get("verify-session", auth=None, url_name="verify_session")
@ensure_csrf_cookie
def verify_session(request):
    if not request.user.is_authenticated:
        return Response({"ok": None}, status=200)
    return Response({"ok": request.user.username}, status=200)


register_router = Router()


@register_router.post("register", auth=None, response={201: None, 400: ErrorOut}, url_name="register_user")
def register(request, user_payload: UserRegisterSchema):
    try:
        validate_password(user_payload.password)
    except ValidationError as e:
        return 400, {"message": e.messages}

    if user_payload.password != user_payload.password_confirm:
        return 400, {"message": "Password does not match"}

    email_validator = EmailValidator()
    try:
        email_validator(user_payload.email)
    except ValidationError as e:
        return 400, {"message": e.messages}

    User = get_user_model()

    email_exist = User.objects.filter(email=user_payload.email).exists()
    if email_exist:
        return 400, {"message": "Email already exist"}

    username_exist = User.objects.filter(username=user_payload.username).exists()
    if username_exist:
        return 400, {"message": "Username already exist"}

    User.objects.create_user(
        username=user_payload.username,
        email=user_payload.email,
        password=user_payload.password
    )

    return 201, None



from datetime import timedelta

import pytest
from django.test import RequestFactory
from django.urls import reverse_lazy

from knightauth.auth import TokenAuthentication
from knightauth.models import AuthToken
from knightauth.signals import token_expired


@pytest.fixture
def setup_user(django_user_model):
    django_user_model.objects.create_user(username='john.doe', email='john.doe@example.com', password='qwerty1200')
    django_user_model.objects.create_user(username='jane.doe', email='jane.doe@example.com', password='qwerty1200')


@pytest.mark.django_db
def test_initial_auth_token_to_be_zero(setup_user):
    assert AuthToken.objects.count() == 0


@pytest.mark.django_db
def test_login_creates_auth_token(setup_user, django_user_model, client):
    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_login'),
        'data': {
            django_user_model.USERNAME_FIELD: "john.doe",
            "password": "qwerty1200"
        },
        'content_type': 'application/json'
    }

    client.post(**request_kwargs)
    client.post(**request_kwargs)
    client.post(**request_kwargs)

    assert AuthToken.objects.count() == 3
    assert all(obj.token_key for obj in AuthToken.objects.all())


@pytest.mark.django_db
def test_login_returns_serialized_token(setup_user, django_user_model, client):
    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_login'),
        'data': {
            django_user_model.USERNAME_FIELD: "john.doe",
            "password": "qwerty1200"
        },
        'content_type': 'application/json'
    }

    response = client.post(**request_kwargs)
    json_response = response.json()

    assert response.status_code == 200
    assert json_response.get('token')


@pytest.mark.django_db
def test_logout_deletes_keys(setup_user, django_user_model, client):
    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_login'),
        'data': {
            django_user_model.USERNAME_FIELD: "john.doe",
            "password": "qwerty1200"
        },
        'content_type': 'application/json'
    }

    client.post(**request_kwargs)
    response = client.post(**request_kwargs).json()

    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_logout'),
        'content_type': 'application/json',
        'HTTP_AUTHORIZATION': response.get('token')
    }

    client.post(**request_kwargs)

    assert AuthToken.objects.count() == 1


@pytest.mark.django_db
def test_logout_all_deletes_keys(setup_user, django_user_model, client):
    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_login'),
        'data': {
            django_user_model.USERNAME_FIELD: "john.doe",
            "password": "qwerty1200"
        },
        'content_type': 'application/json'
    }

    client.post(**request_kwargs)
    response = client.post(**request_kwargs).json()

    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_logoutall'),
        'content_type': 'application/json',
        'HTTP_AUTHORIZATION': response.get('token'),
    }

    client.post(**request_kwargs)

    assert AuthToken.objects.count() == 0


@pytest.mark.django_db
def test_logout_all_deletes_keys_for_user(setup_user, django_user_model, client):
    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_login'),
        'data': {
            django_user_model.USERNAME_FIELD: "john.doe",
            "password": "qwerty1200"
        },
        'content_type': 'application/json'
    }
    client.post(**request_kwargs)

    request_kwargs['data'][django_user_model.USERNAME_FIELD] = "jane.doe"
    client.post(**request_kwargs)
    response = client.post(**request_kwargs).json()

    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:token_logoutall'),
        'content_type': 'application/json',
        'HTTP_AUTHORIZATION': response.get('token'),
    }

    client.post(**request_kwargs)

    assert AuthToken.objects.count() == 1


@pytest.mark.django_db
def test_expired_tokens_login_fails(setup_user, django_user_model, client):
    _, token = AuthToken.objects.create(user=django_user_model.objects.first(), expiry=timedelta(seconds=0))

    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:test'),
        'content_type': 'application/json',
        'HTTP_AUTHORIZATION': token
    }

    response = client.get(**request_kwargs)

    assert response.status_code == 401
    assert response.json().get('detail') == 'Unauthorized'


@pytest.mark.django_db
def test_token_authentication_with_valid_token(setup_user, django_user_model):
    user = django_user_model.objects.first()
    _, token = AuthToken.objects.create(user=user)

    auth_user = TokenAuthentication().authenticate(RequestFactory().get('/'), token)

    assert auth_user == user


@pytest.mark.django_db
def test_token_authentication_with_invalid_token(setup_user):
    rf = RequestFactory()

    assert TokenAuthentication().authenticate(RequestFactory().get('/'), '') is None
    assert TokenAuthentication().authenticate(RequestFactory().get('/'), 'This is token') is None


@pytest.mark.django_db
def test_expiry_signals(setup_user, django_user_model, client):
    signal_fired = False

    def signal_handler(sender, **kwargs):
        nonlocal signal_fired
        signal_fired = True

    token_expired.connect(signal_handler)

    _, token = AuthToken.objects.create(user=django_user_model.objects.first(), expiry=timedelta(seconds=0))

    request_kwargs = {
        'path': reverse_lazy('api-1.0.0:test'),
        'content_type': 'application/json',
        'HTTP_AUTHORIZATION': token
    }

    client.get(**request_kwargs)

    assert signal_fired



from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils import timezone

from knightauth import crypto
from knightauth.settings import CONSTANTS, knight_auth_settings

sha = knight_auth_settings.SECURE_HASH_ALGORITHM

User = get_user_model()


class AuthTokenManager(models.Manager):
    def create(
            self,
            user,
            expiry=knight_auth_settings.TOKEN_TTL,
            prefix=knight_auth_settings.TOKEN_PREFIX
    ):
        token = prefix + crypto.create_token_string()
        digest = crypto.hash_token(token)
        if expiry is not None:
            expiry = timezone.now() + expiry
        instance = super(AuthTokenManager, self).create(
            token_key=token[:CONSTANTS.TOKEN_KEY_LENGTH], digest=digest,
            user=user, expiry=expiry)
        return instance, token


class AbstractAuthToken(models.Model):
    objects = AuthTokenManager()

    digest = models.CharField(
        max_length=CONSTANTS.DIGEST_LENGTH,
        primary_key=True
    )
    token_key = models.CharField(
        max_length=CONSTANTS.MAXIMUM_TOKEN_PREFIX_LENGTH + CONSTANTS.TOKEN_KEY_LENGTH,
        db_index=True
    )
    user = models.ForeignKey(
        User,
        null=False,
        blank=False,
        related_name='auth_token_set',
        on_delete=models.CASCADE
    )
    created = models.DateTimeField(auto_now_add=True)
    expiry = models.DateTimeField(null=True, blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return '%s : %s' % (self.digest, self.user)


class AuthToken(AbstractAuthToken):
    class Meta:
        swappable = 'KNIGHT_AUTH_TOKEN_MODEL'


def get_token_model():
    try:
        return apps.get_model(knight_auth_settings.TOKEN_MODEL, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "TOKEN_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "TOKEN_MODEL refers to model '%s' that has not been installed"
            % knight_auth_settings.TOKEN_MODEL
        )



import binascii
from hmac import compare_digest

from django.utils import timezone
from ninja.security import APIKeyHeader

from knightauth.crypto import hash_token
from knightauth.models import get_token_model
from knightauth.settings import CONSTANTS, knight_auth_settings
from knightauth.signals import token_expired


class TokenAuthentication(APIKeyHeader):
    param_name = "Authorization"

    def authenticate(self, request, token):
        user, auth_token = self.authenticate_credentials(token)
        request._auth = auth_token

        return user

    def authenticate_credentials(self, token):
        if not token:
            return None, None

        auth_tokens = (
            get_token_model()
            .objects
            .filter(token_key=token[:CONSTANTS.TOKEN_KEY_LENGTH])
        )

        for auth_token in auth_tokens:
            if self._cleanup_token(auth_token):
                continue

            try:
                digest = hash_token(token)
            except (TypeError, binascii.Error):
                return None, None

            if compare_digest(digest, auth_token.digest):
                if knight_auth_settings.AUTO_REFRESH and auth_token.expiry:
                    self.renew_token(auth_token)

                return self.validate_user(auth_token)

        return None, None

    def renew_token(self, auth_token):
        current_expiry = auth_token.expiry
        new_expiry = timezone.now() + knight_auth_settings.TOKEN_TTL
        auth_token.expiry = new_expiry
        # Throttle refreshing of token to avoid db writes
        delta = (new_expiry - current_expiry).total_seconds()
        if delta > knight_auth_settings.MIN_REFRESH_INTERVAL:
            auth_token.save(update_fields=('expiry',))

    def validate_user(self, auth_token):
        if not auth_token.user.is_active:
            return None, None

        return auth_token.user, auth_token

    def _cleanup_token(self, auth_token):
        for other_token in auth_token.user.auth_token_set.all():
            if other_token.digest != auth_token.digest and other_token.expiry:
                if other_token.expiry < timezone.now():
                    other_token.delete()
                    username = other_token.user.get_username()
                    token_expired.send(
                        sender=self.__class__,
                        username=username,
                        source="other_token"
                    )

        if auth_token.expiry is not None:
            if auth_token.expiry < timezone.now():
                username = auth_token.user.get_username()
                auth_token.delete()
                token_expired.send(
                    sender=self.__class__,
                    username=username,
                    source="auth_token"
                )
                return True

        return False





from typing import Union

from ninja import Schema
from datetime import datetime


class LoginIn(Schema):
    username: str
    password: str


class ErrorOut(Schema):
    message: Union[str, list]


class LoginSuccessOut(Schema):
    expiry: datetime
    token: str


class UserRegisterSchema(Schema):
    username: str
    email: str
    password: str
    password_confirm: str



# Generated by Django 4.2.4 on 2023-08-12 15:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthToken',
            fields=[
                ('digest', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('token_key', models.CharField(db_index=True, max_length=25)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('expiry', models.DateTimeField(blank=True, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_token_set', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'swappable': 'KNOX_TOKEN_MODEL',
            },
        ),
    ]



import binascii
from os import urandom as generate_bytes

from knightauth.settings import knight_auth_settings

hash_func = knight_auth_settings.SECURE_HASH_ALGORITHM


def create_token_string():
    return binascii.hexlify(
        generate_bytes(int(knight_auth_settings.AUTH_TOKEN_CHARACTER_LENGTH / 2))
    ).decode()


def make_hex_compatible(token: str) -> str:
    return binascii.unhexlify(binascii.hexlify(bytes(token, 'utf-8')))


def hash_token(token: str) -> str:
    digest = hash_func()
    digest.update(make_hex_compatible(token))
    return digest.hexdigest()



from django.contrib import admin

from knightauth import models


@admin.register(models.AuthToken)
class AuthTokenAdmin(admin.ModelAdmin):
    list_display = ('digest', 'user', 'created',)
    fields = ()
    raw_id_fields = ('user',)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="knight-auth",
    version="0.3.0",
    description='Authentication for django-ninja, KNOX inspired',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/knightSarai/knight-auth',
    author='knightSarai',
    author_email='knight.sarai.dev@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.11',
        'Framework :: Django',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
    ],
    packages=find_packages(exclude=['core']),
    python_requires='>=3.11.0',
    install_requires=[
        'django>=4.2.4',
        'django-ninja>=1.0'
    ],
    extras_require={
        "dev": ["twine>=4.0.2", "pytest-django>=4.5.2"],
    }
)



from datetime import timedelta

from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string


USER_SETTINGS = getattr(settings, 'KNIGHT_AUTH', {})

IMPORT_STRINGS = [
    'SECURE_HASH_ALGORITHM',
    'USER_SERIALIZER',
]

ISO_8601 = 'iso-8601'

DEFAULTS = {
    'SECURE_HASH_ALGORITHM': 'hashlib.sha512',
    'AUTH_TOKEN_CHARACTER_LENGTH': 64,
    'TOKEN_TTL': timedelta(hours=10),
    'USER_SERIALIZER': None,
    'TOKEN_LIMIT_PER_USER': None,
    'AUTO_REFRESH': False,
    'MIN_REFRESH_INTERVAL': 60,
    'AUTH_HEADER_PREFIX': 'Token',
    'EXPIRY_DATETIME_FORMAT': ISO_8601,
    'TOKEN_MODEL': getattr(settings, 'KNIGHT_AUTH_MODEL', 'knightauth.AuthToken'),
    'TOKEN_PREFIX': '',
}


def perform_import(val, setting_name):
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class APISettings:
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'KNIGHT_AUTH', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        if attr in self.import_strings:
            val = perform_import(val, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


knight_auth_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*args, **kwargs):
    global knight_auth_settings
    setting, value = kwargs['setting'], kwargs['value']
    if setting == 'KNIGHT_AUTH':
        knight_auth_settings = APISettings(**value)
        if len(knight_auth_settings.TOKEN_PREFIX) > CONSTANTS.MAXIMUM_TOKEN_PREFIX_LENGTH:
            raise ValueError("Illegal TOKEN_PREFIX length")


setting_changed.connect(reload_api_settings)


class CONSTANTS:
    TOKEN_KEY_LENGTH = 15
    DIGEST_LENGTH = 128
    MAXIMUM_TOKEN_PREFIX_LENGTH = 10

    def __setattr__(self, *args, **kwargs):
        raise Exception('''
            Constant values must NEVER be changed at runtime, as they are
            integral to the structure of database tables
            ''')


CONSTANTS = CONSTANTS()
